/*
 *  Widget to present an image of the monitors on the Desktop/X11-screen
 */

/*
 * Copyright 2011 - 2019 Zytronic Displays Limited, UK.
 *
 * This file is part of the ZyConfig application.
 *
 * ZyConfig is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * ZyConfig is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with ZyConfig.
 *
 * If not, see <https://www.gnu.org/licenses/>.
 */

#include "DesktopImage.h"

static const int IMG_SIZE = 300;

DesktopImage::DesktopImage ( QWidget *parent ) : QWidget ( parent )
{
    QSizePolicy policy ( QSizePolicy::Preferred, QSizePolicy::Preferred );
    policy.setHeightForWidth ( true );
    setSizePolicy( policy );
    outline = NULL;
    monitorVector = NULL;
    selectedMon = NULL;
}

int DesktopImage::heightForWidth (int width) const
{
    return width;
}

QSize DesktopImage::sizeHint() const
{
    return QSize(IMG_SIZE+2, IMG_SIZE+2);
}

void DesktopImage::setOutline( Monitor * m )
{
    // copy pointer to parent's outline struct
    outline = m;
}

void DesktopImage::setSelected( Monitor * s )
{
    selectedMon = s;
}

void DesktopImage::addMonitors( QVector<Monitor*> * monVec )
{
    monitorVector = monVec;
}

void DesktopImage::paintEvent(QPaintEvent * /*pev*/ )
{
    QPainter p(this);

    if (outline == NULL) return;

    Size2d sz = outline->getSize();
    double scale = ( sz.x > sz.y ) ? sz.x : sz.y;
    int wh2 = ( sz.x - sz.y ) / 2;
    int xoff = 2 * scale / IMG_SIZE, yoff = 16 * scale /IMG_SIZE;     // Text Offset after scaling
    QPen pen = QPen();

    p.translate( 1, 1 );
    p.scale(IMG_SIZE/scale, IMG_SIZE/scale);

    Location loc = { 1, 1 } ;
    if (wh2 > 0) {
        loc.y += wh2;
    } else {
        loc.x -= wh2;
    }
    p.fillRect ( loc.x,loc.y,  sz.x,sz.y, Qt::white );

    // add the individual monitors
    if (monitorVector != NULL)
    {
        QFont font = p.font();
        font.setPixelSize(14*scale/IMG_SIZE);
        p.setFont(font);
        pen.setStyle( Qt::DashLine );
        pen.setWidth(scale/IMG_SIZE);
        pen.setColor( Qt::green );
        p.setPen( pen );
        Monitor *aMon;
        foreach ( aMon, *monitorVector )
        {
            Size2d sz = aMon->getSize();
            Location loc = aMon->getLocation();
            if (wh2 > 0) {
                loc.y += wh2;
            } else {
                loc.x -= wh2;
            }

            p.drawRect( loc.x, loc.y, sz.x, sz.y  );
            p.drawText( loc.x+xoff, loc.y+yoff, aMon->getName(true) );
        }
    }

    // highlight the selected rectangle
    pen.setStyle( Qt::SolidLine );
    pen.setColor( Qt::blue );
    p.setPen( pen );
    if (selectedMon != NULL)
    {
        Size2d sz = selectedMon->getSize();
        Location loc = selectedMon->getLocation();
        if (wh2 > 0) {
            loc.y += wh2;
        } else {
            loc.x -= wh2;
        }
        p.drawRect( loc.x, loc.y, sz.x, sz.y  );
        p.drawText( loc.x+xoff, loc.y+yoff, selectedMon->getName(true) );
    }
}

